#!/usr/bin/env python3
# Adapted from Eli Bendersky's optimized BF interpreter.

import sys      # for accessing command-line args

MEMORY_SIZE = 30000

# -----------------------------------------------------
# This function parses a program string,
#      filtering out all non-command characters.
# -----------------------------------------------------

def parse_bf(src_str):
    program = ''
    for c in src_str:
        if c in '><+-.,[]':
            program += c
    return program

# -----------------------------------------------------
# This function precomputes matching brackets that open
#      and close loops, so that the interpreter can
#      jump immediately at run-time.
# -----------------------------------------------------

def compute_jumptable(program):
    pc = 0
    program_size = len(program)
    jumptable = [0] * program_size
    while (pc < program_size):
        instruction = program[pc]
        if instruction == '[':
            bracket_nesting = 1
            seek = pc
            while bracket_nesting > 0 and seek < (program_size - 1):
                seek += 1
                if program[seek] == ']':
                    bracket_nesting -= 1
                elif program[seek] == '[':
                    bracket_nesting += 1
            if bracket_nesting == 0:
                jumptable[pc] = seek
                jumptable[seek] = pc
            else:
                raise ValueError('unmatched [ at pc = ', pc)
        pc += 1
    return jumptable
                
# -----------------------------------------------------
# The interpreter
# -----------------------------------------------------

def interpret(program, jumptable):

    memory   = [0] * MEMORY_SIZE
    pc       = 0
    data_ptr = 0

    while pc < len(program):
        instruction = program[pc]

        if instruction == '>':
            data_ptr += 1
        elif instruction == '<':
            data_ptr -= 1
        elif instruction == '+':
            memory[data_ptr] += 1
        elif instruction == '-':
            memory[data_ptr] -= 1
        elif instruction == '.':
            print(chr(memory[data_ptr]), end='')
        elif instruction == ',':
            memory[data_ptr] = ord(sys.stdin.read(1))
        elif instruction == '[':
            if memory[data_ptr] == 0:
                pc = jumptable[pc]
        elif instruction == ']':
            if memory[data_ptr] != 0:
                pc = jumptable[pc]
        else:
            msg = 'bad character {} at pc = {}'.format(instruction, pc)
            raise ValueError(msg)

        pc += 1

# -----------------------------------------------------
# The main function takes the command-line args as an
# array.  args[1] is the filename.
# -----------------------------------------------------

def main(args):
    bf_file_path = args[1]

    srcfile = open(bf_file_path, 'r')
    program = parse_bf(srcfile.read())

    # before running the program, pre-compute the jump targets
    jumptable = compute_jumptable(program)

    interpret(program, jumptable)

# -----------------------------------------------------
# These lines call the main function with one of three
# test programs, or with the command-line args
# -----------------------------------------------------

if __name__ == "__main__":
    # program_name = 'hello-world.bf'  # 'cell-size.bf' 'factor.bf'
    # main(['optimized', program_name])
    main(sys.argv)
